#include <string>
#include <cstdio>
#include <cstdlib>
#include <iostream>
using namespace std;

#include "VDIF_Frame.hpp"
#include "Raw_Proto.hpp"
#include "StaticVals.hpp"

const long ELECS_NO=102300;

/** General Configulation data */
Config_set vdif_conf;

/**! "main_cmn.hpp": Common Declraration of static parameters. 
 * Should be included in main file only. */
#include "main_cmn.hpp"
const char* version = "$Revision: 273 $ -Latest Update: $Date: 2017-11-15 10:02:19 +0900 (水, 15 11月 2017) $";


void usage(){

  cerr << "Usage: " <<  me << " -f <filename> [-o <outdir>] [-n <nframe/sec>] [-i <stnid>] ";
  cerr << "         [-l <sec>] [-s <stname>] [-v < -2 - 2 >] \n";
  cerr << "  Converting Elecs VDIF file to normal VDIF. \n";
  cerr << "  -f : input Elecs file.\n";
  cerr << "  -o : Output VDIF file.\n";
  cerr << "  -n : Number of frames in one second.\n";
  cerr << "  -i : Station ID recordid in VDIF header.\n";
  cerr << "  -s : Station Name recordid in EDV(ver.1) header.\n";
  cerr << "  -l : length of conversion (sec). Default=all data.\n";
  cerr << "  -v: " << Verb_str() << endl;
  cerr << "  Ver.:" << version << endl;
  exit(1);
}

int main(int argc, char *argv[]){
  SetMyname(argv[0]);
  string infname, outfile;
  int mode=0;
  int k=0;
  string stname, stid;
  long nf_sec=0;
  int lsec=0;

  while (k >= 0) { 
    switch (k = getopt(argc, argv, "hf:o:i:n:v:l:s:")) { 
    case 'h':  usage(); break;
    case 'f':    infname = optarg;      break;
    case 'o':    outfile = optarg;      break;
    case 's':    stname = optarg;      break;
    case 'n':    nf_sec = strtol(optarg,0,10);      break;
    case 'l':    lsec = strtol(optarg,0,10);      break;
    case 'i':    stid = optarg;      break;
    case 'v':    verbose_no=  (Verbose_level_t)strtol(optarg,0,10);      break;
    case 'd':    mode = strtol(optarg,0,10);      break;
    case '?' : /* Error:  character not in list */ 
      fprintf(stderr, /* Yes */ 
	      "%s WARNING:  Unknown option on command line \n", me); 
      usage();
      break; 
    case ':' : /* Error:  missing parameter */ 
      fprintf(stderr, "%s WARNING:  Missing number on command line \n", me); 
      usage();
      break; 
    }
  }

  if(infname.empty() || outfile.empty()) usage();

  double total_wsize=0.0;
  long total_wframe=0;
  struct timeval tv0, tv1;  

  gettimeofday(&tv0,0); // Processing start time
    
  VDIF_Frame  vdif;
  if( ! vdif.ROpen(infname)){
    string msg =  "Error in opening file: " +  infname;
    throw msg;
  }

  uint8_t edv=0;
  if( ! vdif.find_hdr()  ){
    fprintf(stderr,"Error: Failed to find frame.\n");
    exit(1);
  }else{
    fprintf(stderr,"Hdr-Found:Offset=%d \n",vdif.Hdr_Offset() );
    edv=vdif.Hdr().EDV(); 
  };
  

  size_t framesize = vdif.Framesize();
  size_t hdrsize = vdif.Headsize();
  long datsize = framesize - hdrsize;
  int nch = vdif.Hdr().Nch();
  int qbit = vdif.Hdr().Qbit();
  //  long nframe_sec = vdif.Nframe_sec_estimate();
  long nframe_sec = vdif.Nframe_sec();
  double srate = (double) datsize *vdif.Nframe_sec() * 8/qbit/nch*1.e-6; // Msps
    
  VDIF_Frame new_vdif = vdif;
  if(nf_sec > 0)  new_vdif.Set_Nframe_sec(nf_sec);
  new_vdif.Set_Sparam(nch,qbit,(uint32_t)srate*1000000);
  size_t new_framesize = new_vdif.Framesize();
  long data_sec = nframe_sec*datsize;

  uint8_t *databuf=new uint8_t[data_sec];


  try{
    if(! stid.empty()) new_vdif.Set_StnID(stid.c_str());
    if(! stname.empty()) new_vdif.Set_StnameID(stname.c_str());

    fprintf(stderr,"----------------------------------------\n");
    fprintf(stderr,"Framesize: %5ld [byte] data/frame: %5ld [byte]", framesize, datsize);
    fprintf(stderr,"  Nch  : %4d   Qbit : %2d\n", nch, qbit);
    fprintf(stderr,"Estimated Frame/sec : %6ld  Data/sec %4ld [MByte/sec]\n", 
	    nframe_sec, data_sec/1000000 );
    fprintf(stderr,"----------------------------------------\n");

    cerr << "========= Conv Elevs2VDIF ver. 1.$Rev: 273 $ start ================\n";
    fprintf(stdout,"=== Converting frame size from %8u Byte (%8u frames/sec) ", 
	    (unsigned int)framesize, (unsigned int)vdif.Nframe_sec());
    fprintf(stdout,"to %8u Byte (%8u frames/sec) \n",
	    (unsigned int)new_framesize, (unsigned int)nf_sec); 


    RAW_Proto file_prot;
    file_prot.Set_VlbiFrame(&new_vdif);
    //    DAS_Info das((uint32_t)(srate*1.e6),nch,qbit);
    //    file_prot.Set_DasInf(das);
    file_prot.Set_Host(outfile.c_str());
    if(file_prot.Tx_Ready()){
      printf(" Converting from (%s) to (%s) \n", 
	     infname.c_str(), outfile.c_str());
    }else{
      char buf[2048];
      string msg = "E Failed to open outfile: " + outfile + ":";
      msg += strerror_r(errno,buf,2048);
      throw msg;
    }
    
    uint8_t *pbuf = databuf;
    uint8_t *frame_buf = new uint8_t [framesize+128];
    uint8_t *pdat = frame_buf +hdrsize;
    VDIF_Hdr * phdr = (VDIF_Hdr *)frame_buf;
    int nread=0;
    
    long cnt=0;
    long tcnt=0;
    Time_dat t1=vdif.Hdr().Epoch();
    size_t wsize=0;
    int tsec=0;
    while( ! vdif.EndofFile_Yes() ){
      if(tcnt == ELECS_NO ) {
	nread = vdif.Read_Bytes(frame_buf, framesize+128);
	tcnt=0;
      }else
	nread = vdif.Read_Bytes(frame_buf, framesize);

      tcnt++;
      
      if (edv == phdr->EDV() && phdr->HD_Match()  ){
	if( t1 != phdr->Epoch() ){
	  cout << phdr->Str_hd(VDIF_Hdr::ONELINE) << " " << cnt << endl;
	  new_vdif.Set_Epoch(t1.Time_t());
	  wsize = file_prot.TX1sec(t1.Time_t(), databuf, data_sec);
	  total_wsize += (double)wsize;
	  total_wframe+= new_vdif.Nframe_sec();
	  
	  fprintf(stderr,"wrote  %.0lf MBytes. \r", (total_wsize*1.e-6)); 	  
	  t1 = phdr->Epoch();
	  bzero(databuf,data_sec);
	  cnt=0;
	  tsec++;
	}
	
	if( nread > 0 ){
	  pbuf = databuf+ phdr->iFrame() * datsize;
	  memcpy(pbuf,pdat,nread-hdrsize); 
	}
      }

      cnt++;
      if(lsec > 0 && tsec > lsec) break;
    } // while 

  delete [] frame_buf;
  }// End of Try 

  catch(string msg){
    cerr << msg << endl;
  }

  delete [] databuf;

  

  gettimeofday(&tv1,0); // Processing End time
  double dt = (tv1.tv_sec - tv0.tv_sec)+  (tv1.tv_usec - tv0.tv_usec)*1.e-6 ;
  double wrate = total_wsize*1.e-6/dt; // MB/sec
  fprintf(stderr,"wrote  %.1lf MBytes. %ld frames\n", total_wsize*1.e-6, total_wframe); 
  fprintf(stderr," %.3lf sec elapsed(%.1lf MB/s). Finished.\n", dt, wrate);
  cerr << "==========================================\n";
  

}
