
/** @file vdif_chk is test reading VDIF header data 
 * @Version $Rev: 273 $
 * @Author M.Sekido
 * @Date  $Date: 2017-11-15 10:02:19 +0900 (水, 15 11月 2017) $
 * $Id: vdif_chk.cpp 272 2017-11-15 01:00:12Z vlbi $ 
 */

const char* version = "$Revision: 273 $ -Latest Update: $Date: 2017-11-15 10:02:19 +0900 (水, 15 11月 2017) $";

#include <iostream>
#include <string>
#include <vector>
using namespace std;

#include <cstdio>
#include <cstdlib>

#include "StaticVals.hpp"
#include "VDIF_Header.hpp"
#include "VDIF_Frame.hpp"
#include "DAS_info.hpp"
#include "VsiData.hpp"
#include "hist_g.hpp"
#include "conf_utl.hpp"

extern "C"{
#include <unistd.h>
}

/**! "main_cmn.hpp": Common Declraration of static parameters. 
 * Should be included in main file only. */
#include "main_cmn.hpp"

/** General Configulation data */
Config_set vdif_conf;

void usage(){
  cerr << "Usage: " << 
    me << " -f <filename> [-m <e|s|h>] [-l <sec> ][-n <int>] [-v <n>] [-d <n>][-g]\n";
  cerr << "        [-b <bit encode>] [-s <nframe>] [-y <year>] [-t <text out>]\n";
  cerr << "      [-e <yyyy.ddd.HH:MM:SS>] [-m <Frame Slip Margin>]\n";
  cerr << "-l : Length of sec for read. (default=1 sec.), 0 means reading to the end.\n";
  cerr << "-n : Number of frames to be displayed. (default:10), 0:all data.\n";
  cerr << "-m : Display Mode, (e)very-frame, every-(s)econd, or (h)istgram \n";
  cerr << "-d : Display info 0:simple(defalt),\n";
  cerr << "     1:EDVer RefEpoch,2:StnID,Thread No.,DataType,\n";
  cerr << "     3:same with 2 but one line, 4:Header with data.\n";
  cerr << "     -1:supress Warning, -2:Supress Info.\n";
  cerr << "-s : Skip First <n> frames, in case of every frame(-d) option.\n";
  cerr << "-y : Year used for matching."  << endl;
  cerr << "-t : Extracting data to file."  << endl;
  cerr << "-e : Epoch to extract data."  << endl;
  cerr << "-z : 'Frame slip margin' is # of Frames allowed in framenumber jump and disordering.\n";
  cerr << "     This is used in VDIF_Frame::Nframe_sec_estimate() to determin datarate."<< endl;
  cerr << "-v : " << Verb_str() <<" Default=-3:No output." << endl;
  cerr << "-g : Histgram Display in case of every frame(-d).\n";
  //  cerr << "-e : Extended version Specification (NICT or none). Default is none\n";
  cerr << "-b  : Choice of bit encoding for histgram.\n";
  cerr << "      one of [ADS2000/ADS3000/SignMag/RevSign].(default=ADS2000)\n";
  cerr << "           ADS2000: 00,01,10,11 from bottom to top level.\n";
  cerr << "           ADS3000: 00,10,01,11 from bottom to top level.\n";
  cerr << "           SignMag: 01,00,10,11 from bottom to top level.\n";
  cerr << "           RevSgn : 10,00,01,11 from bottom to top level.\n";
  cerr << "  Ver.:" << version << endl;
  exit(1);
}

int skip_epoch(Time_dat &p_epoch, VDIF_Frame& vdif){

  int nframe_sec = vdif.Nframe_sec();
  while( !vdif.EndofFile_Yes() && vdif.Hdr().Epoch() < p_epoch){
    if(vdif.Hdr().iFrame() < nframe_sec / 2 ) vdif.SeekFrames(nframe_sec/2);
    else vdif.SeekFrames(1);
    vdif.find_hdr();
    //	vdif.Read_Bytes(buf,frmsize);
  }
  if(vdif.EndofFile_Yes()){
    cerr << "Eof Found.\n"; exit(1);
  }
  fprintf(stderr,"Data Time:%s \n",
	  vdif.Hdr().Epoch().Print_time(14).c_str());
  //    delete [] buf;

  if( vdif.Hdr().Epoch() != p_epoch)  {
    fprintf(stderr,"E Failed to find epch(%s)\n", 
	    p_epoch.Print_time(14).c_str());
    exit(1);
  }
  return 0;
}

int main(int argc, char *argv[]){
  me = (me = strrchr(argv[0], '/')) == NULL ? argv[0] : me+1; /* My name */

  string fname;
  //  VDIF_Ext_t ext_t = No_ext; // No VDIF-extended version
  int ndisp = 10;
  int mode = 1;  // 0: Every frame, 1: every second,
  const int Ncmp = 100000;
  int year=0;
  //  bool histgram_on = false;
  int nskip_frame(0);
  int disp_num=0;
  Time_dat t_epoch;  bool epoch_flag=false; 
  int nsec=0;
  string text_out;
  VDIF_Frame  vdif;
  VsiData::BitEnc_t code_type = VsiData::ADS2000_2bit;
  VDIF_Hdr::Prnt_t idsp(VDIF_Hdr::ONELINE);
  verbose_no= (Verbose_level_t)-3; // No print info
  int k=0;
  while (k >= 0) { 
    switch (k = getopt(argc, argv, "hf:b:m:n:v:s:l:y:t:d:e:z:")) { 
    case 'h':  usage(); break;
    case 'f':    fname = optarg;      break;
    case 't':    text_out = optarg;      break;
    case 'y':    year = strtol(optarg,0,10);      break;
    case 's':    nskip_frame = strtol(optarg,0,10);      break;
    case 'l':    nsec = strtol(optarg,0,10);      break;
    case 'z':   {float  tmpnum = strtof(optarg,0); vdif.Set_FrameSlipMargin(tmpnum); } break; 
    case 'e':   {string tmpbuf = optarg;  epoch_flag=true; 
	t_epoch.Set_time(tmpbuf,5,":.");  } break;
      //    case 'g':    histgram_on = true;      break;
    case 'm':    
      switch (optarg[0]){
      case 'e' : mode=0;      break; // Every frame
      case 's' : mode=1;      break; // Every second
      case 'h' : mode=2;      break; // Histgram
      case 't' : mode=3;      break; // text output
      } break;
    case 'v':  verbose_no = (Verbose_level_t)strtol(optarg,0,10);      break;
    case 'd':  disp_num = strtol(optarg,0,10);  break;
    case 'b':
      if( strncmp(optarg, "ADS3000",7)==0) code_type = VsiData::ADS3000_2bit;
      if( strncmp(optarg, "SignMag",7)==0) code_type = VsiData::SignMag_2bit;
      if( strncmp(optarg, "RevSign",7)==0) code_type = VsiData::RevSgnMag_2bit;
      break;
    case 'n':    ndisp = atoi(optarg);      break;
      // case 'e':    if(strcmp(optarg,"NICT")==0) ext_t = NICT_ext; break;
      //    case 'l':    dlevel = atoi(optarg);      break;
      //    case 'd': vdif_hdr.Set_dbg(atoi(optarg)); break;
    case '?' : /* Error:  character not in list */ 
      fprintf(stderr, /* Yes */ 
	      "%s WARNING:  Unknown option on command line \n", me); 
      usage();
      break; 
    case ':' : /* Error:  missing parameter */ 
      fprintf(stderr, "%s WARNING:  Missing number on command line \n", me); 
      usage();
      break; 
    }
  }

  if(! text_out.empty()){
    mode = 3; // Text ouput 
    ndisp=2048;
  }
    
  switch(disp_num){
  case 1: idsp = VDIF_Hdr::ALL_ONE; break;
  case 2:idsp = VDIF_Hdr::ALL; break;
  case 3:idsp = VDIF_Hdr::CONST; break;
  case 4:idsp = VDIF_Hdr::ALL_ONE; break;
  case 5:idsp = VDIF_Hdr::RAW; break;
  default:
    idsp = VDIF_Hdr::ONELINE; break;
  }

  if(fname.empty()) usage();
  //  cout << "=== " << version << "===" << endl;  
  //  FILE *fp = fopen(fname.c_str(),"r");

  if( ! vdif.ROpen(fname)){
    cerr << "Error in opening file: " << fname << endl;
    exit(1);
  }
 
  if( !vdif.find_hdr() ){
    if(! vdif.EndofFile_Yes() )  cerr << "E Failed to find VDIF Header.\n";
    exit(1);
  }else{
    fprintf(stderr,"Hdr-Found:Offset=%d \n",vdif.Hdr_Offset() );
  }

  // Skip frames until specified Epoch 
  if(epoch_flag) skip_epoch(t_epoch,vdif);

  long framesize = vdif.Framesize();
  long datsize = framesize - vdif.Headsize();
  int nch = vdif.Hdr().Nch();
  int qbit = vdif.Hdr().Qbit();
  //  long nframe_sec = vdif.Nframe_sec_estimate();
  long nframe_sec = vdif.Nframe_sec();
  long srate = vdif.Hdr().Srate();
  
  double drate = (double)srate/1000000*nch*qbit/8; // MB/sec    
  if(vdif.Hdr().EDV() != EDV_NICT){
    drate = (double)nframe_sec*datsize;
    srate = 8*datsize*nframe_sec/nch/qbit;
  }
  DAS_Info das = vdif.DAS_Inf();// Sampling info
  das.Set_Param(nch,qbit,srate);
  VsiData vsidat(&das);

  fprintf(stdout,"Input File=%s, Inital header is as follows:\n",
	  fname.c_str());
  string strbar(60,'=');
  cerr << strbar << endl;
  vdif.Hdr().Print_hd(stdout);
  fprintf(stdout,"DataRate:%.3lf MB/sec\n", drate);
  fprintf(stdout,"Nframe/sec=%lu, framesize=%ld (datasize:%ld)Byte, current iframe=%ld \n",
	  (long unsigned int)vdif.Nframe_sec(), 
	  (long unsigned int)vdif.Framesize(),datsize,
	  (long int)vdif.Hdr().iFrame());
  
  ReturnFrameInfo retinfo;
  vector <uint8_t> rbuf;
  cerr << strbar << endl;

  vector <VDIF_Hdr> hdrarr(0);
  switch (mode) {
  case 0:{ // Every frame
    // hdrarr.resize(ndisp);
    // int n_frames = vdif.Read_every_hdr(hdrarr,ndisp);
    // for(int i=0; i<n_frames; i++){
    //   string Info1_dmy = hdrarr[i].Str_hd(idsp);
    //   //      if(Info1 != Info1_dmy) 
    //   fprintf(stdout, "%s",Info1_dmy.c_str()); 
    //   fprintf(stdout,"\n");
    // }

    char buf[framesize];
    VDIF_Hdr* phdr = (VDIF_Hdr*)buf;
    for(int i=0; i<ndisp; i++){
      vdif.Read_Bytes(buf,framesize);
      if( !phdr->HD_Match()){
	fprintf(stderr, "1=%d W Header matching failed.\n",i);
	if( !vdif.find_hdr()) break;
      }
      cout << phdr->Str_hd(idsp) << endl;
    }
    
    
    
  }
    break;
  case 1:{  // every second
    //    vdif.Rewind(VDIF_Frame::HEADER);
    int nfrm=10;
    int isec=-1;
    while(! vdif.EndofFile_Yes() && isec < nsec ){
      long nret = vdif.Read_1sec_hdr(hdrarr, nfrm, year);
      fprintf(stdout, "%s %7ld frame/sec.\n", 
	      hdrarr[0].Str_hd(idsp).c_str(), nret); 
      if(nsec >0) isec++;
      //	  fflush(stdout);
    }
  }
    break;
  case 2:{  // histgram
    //    long nframe = 
    vdif.Read_every_sec_hdr(hdrarr, ndisp);
    //    fprintf(stdout,"Total data length: %lu sec \n", hdrarr.size()); 
    // fprintf(stdout," Reading Nsec =%ld sec, %lu frames, ",  
    // 	    hdrarr.size(), (unsigned long)nframe); 
    // fprintf(stdout," Rate(Estimate):%5d frm/sec, Srate/ch: %.3f Msps\n", 
    // 	    vdif.Nframe_sec(), srate);
    
    int npt_frame = datsize * 8/qbit/nch; // #point per frame;
    const int Nframe_comp = npt_frame > Ncmp ? Ncmp : npt_frame;
    int lbound(0), ndiv(2), istep(0);
    switch(qbit){
    case 1: lbound = 0, istep=2; ndiv=2; break;
    case 2: lbound =-2; istep=2; ndiv=4;  break;
    case 4: lbound =-7, istep=1; ndiv=16; break;
    case 8: lbound =-127, istep=1; ndiv=256; break;
    }
    Hist_g <int8_t> hist( lbound, ndiv,istep);
    
    //    vdif.Rewind(VDIF_Frame::HEADER);

    vsidat.Make_Table(code_type); // Make Decode Table 
    if(nskip_frame > 0) vdif.SeekFrames(nskip_frame);
    while(! vdif.EndofFile_Yes() && ndisp > 0 ){
      retinfo = vdif.Read_frame(hdrarr, rbuf, 1);
      if(retinfo.state == ReturnFrameInfo::NG ){
	fprintf(stdout, "E Failed to read frame\n"); break;
      }
      string Info1_dmy = hdrarr[0].Str_hd(idsp);
      //      if(Info1 != Info1_dmy) 
      fprintf(stdout, "%s",Info1_dmy.c_str()); 
	  
      vsidat.Interpret(rbuf,Nframe_comp, 1); // Integer conv.
      if(nch > 7)fputc('\n',stdout);
      fprintf(stdout,"Hist[%%] ");
      for(int ich =0; ich<nch; ich++){
	hist.Init();
	hist.Accm(vsidat.ValArr(ich));
	string str = hist.StrRatioArr(); 
	fprintf(stdout,"(%2dch):%s ", ich, str.c_str());
	if( (ich % 4 == 3) && ich != nch-1) fprintf(stdout,"\n");
      }
      fprintf(stdout,"\n");
      ndisp--;
    } // while
  } 
    break;
  case 3:{ // text output 
    //    vdif.Rewind(VDIF_Frame::HEADER);
    //    double dt = 1.0/nframe_sec;
    Time_dat t0=vdif.Hdr().Epoch();
    const float* pf[nch] ;
    if(vdif.Hdr().Data_Type() == VDIF_Hdr::Complex_Sample){
      switch(qbit){
      case 1:  code_type = VsiData::IQ_1bit;     break;
      case 2:
	code_type = VsiData::IQ_2bit;     break;
      }
    }
    vsidat.Make_Table(code_type); // Make Decode Table 

    FILE *fp_out=0;
    if( (fp_out = fopen(text_out.c_str(),"w")) == NULL){
      cerr << "Failed to open outfile(" << text_out << ")\n";
      exit(1);
    }
    fprintf(fp_out, "# Time CH: ");
    for(int i=0; i<nch; i++) fprintf(fp_out," %d", i+1);
    putc('\n',fp_out);
    
    int ipt=ndisp;
    int npt_frame= datsize*8/nch/qbit;
    int ipt_frame=npt_frame;
    int icnt=0;
    while( ipt > 0 && !vdif.EndofFile_Yes()){
      if(ipt_frame > npt_frame-1){
	retinfo = vdif.Read_frame(hdrarr, rbuf);      
	ipt_frame=0;
	int ncmp = npt_frame > ipt ? ipt : npt_frame;
	vsidat.Interpret( rbuf, ncmp); // Integer conv.
	for(int ich=0; ich<nch; ich++)
	  pf[ich] = vsidat.pfValArr(ich); // Pointer to data
      }
      //      Time_dat t1 = t0 + dt * hdrarr[0].iFrame();
      fprintf(fp_out,"%s %6d %4d",
	      vdif.Hdr().Epoch().Print_time(15).c_str(), 
	      vdif.Hdr().iFrame(), icnt);
      if(icnt<10) fprintf(stdout,"%s %6d %4d",
			  vdif.Hdr().Epoch().Print_time(15).c_str(), 
			  vdif.Hdr().iFrame(),  icnt);

      for(int ich=0; ich<nch; ich++){
	fprintf( fp_out," %+3.1f",  *pf[ich] ); 
	if(icnt<10) fprintf( stdout," %+3.1f",  *pf[ich] ) ; 
	pf[ich]++;
      }
      putc('\n',fp_out);
      if(icnt<10)  putc('\n',stdout) ;
      ipt--;
      ipt_frame++;
      icnt++;
    }
    fclose(fp_out);
  }
    break;
  } // End case 
  
  cerr << strbar << endl;    

    return 0;
}


