#include <string>
#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <boost/algorithm/string.hpp>
using namespace std;

#include "VDIF_Frame.hpp"
#include "Raw_Proto.hpp"
#include "StaticVals.hpp"

/** General Configulation data */
Config_set vdif_conf;

/**! "main_cmn.hpp": Common Declraration of static parameters. 
 * Should be included in main file only. */
#include "main_cmn.hpp"
const char* version = "$Revision: 274 $ -Latest Update: $Date: 2017-11-15 13:14:25 +0900 (水, 15 11月 2017) $";

void usage(){

  cerr << "Usage: " <<  me << " -f <filename>:<filename>:....:<filename> [-o <outdir>] \n ";
  cerr << "        [-n <nframe/sec>] [-i <stnid>] [-l <sec>] [-s <stname>] [-v < -2 - 2 >] \n";
  cerr << "        [-c <0:0:...:0>] \n";
  cerr << " Merging 8 vdif file to single thread vdif file. \n";
  cerr << "  1-8ch of file1-4 and 9-16ch of file5-8 are marged into one file with 64 channels.\n";
  cerr << "   First or last half set of channels of 8 files are marged into one file.\n";
  cerr << "  -f : input files. 8 files selerated by ':'. \n";
  cerr << "  -o : Output VDIF file.\n";
  cerr << "  -n : Number of frames in one second.\n";
  cerr << "  -i : Station ID recordid in VDIF header.\n";
  cerr << "  -s : Station Name recordid in EDV(ver.1) header.\n";
  cerr << "  -l : length of conversion (sec). Default=all data.\n";
  cerr << "  -v: " << Verb_str() << endl;
  cerr << "  -c : 0:First(default) or 1:Last 8 ch of data is copied. Defaut=0:0:0:0:1:1:1:1. \n";
  cerr << "  Ver.:" << version << endl;
  exit(1);
}


bool Sync_frame(vector <VDIF_Frame>& vdif_arr); ///< Synchronize epoch and _fp
bool Check_Homo(vector <VDIF_Frame>& vdif_arr); ///< Check Homogenity 
bool Check_eof(vector <VDIF_Frame>& vdif_arr);

const int N_mrgfile = 8;
// extern char Rot_bar::str[];

int main(int argc, char *argv[]){
  SetMyname(argv[0]);
  string infname, outfile;
  int mode=0;
  int k=0, ih[8]={0,0,0,0,1,1,1,1};
  string stname, stid, instr, ihstr;
  int verbose_no = -1; // No output
  long nf_sec=0;
  int lsec=0;

  while (k >= 0) { 
    switch (k = getopt(argc, argv, "hf:o:i:n:v:l:s:c:")) { 
    case 'h':  usage(); break;
    case 'f':    instr = optarg;      break;
    case 'o':    outfile = optarg;      break;
    case 's':    stname = optarg;      break;
    case 'n':    nf_sec = strtol(optarg,0,10);      break;
    case 'c':    ihstr = optarg;   break;
    case 'l':    lsec = strtol(optarg,0,10)-1;      break;
    case 'i':    stid = optarg;      break;
    case 'v':    verbose_no=  (Verbose_level_t)strtol(optarg,0,10);      break;
    case 'd':    mode = strtol(optarg,0,10);      break;
    case '?' : /* Error:  character not in list */ 
      fprintf(stderr, /* Yes */ 
	      "%s WARNING:  Unknown option on command line \n", me); 
      usage();
      break; 
    case ':' : /* Error:  missing parameter */ 
      fprintf(stderr, "%s WARNING:  Missing number on command line \n", me); 
      usage();
      break; 
    }
  }

  if(instr.empty() || outfile.empty()) usage();
  vector <string> infiles;  
  boost::algorithm::split(infiles, instr, boost::is_any_of(":"));
  if( infiles.size() < 8){
    cerr << " input file must be 8.\n"; 
    usage();
  }
  if( ! ihstr.empty() ){
    vector <string> token;
    boost::algorithm::split(token, ihstr, boost::is_any_of(":"));
    if( token.size() < 8) {
      fprintf(stderr, "Error: -c flag must be 8., but string=%s (%u) \n",
	      ihstr.c_str(),(unsigned int)token.size());
      usage();
    }else{
      for(int i=0;i<8; i++) ih[i] = strtol(token[i].c_str(),0,10);
    }
  }
  
  char hchstr[2][20] = { "0-8", "9-15"};
  vector <VDIF_Frame>  in_vdif( 8 );
 
  uint8_t edv=0;
  for(int i=0; i < 8 ; i++){
    if( ! in_vdif[i].ROpen(infiles[i]) ) {
      fprintf(stderr, "E main(): Failed to open vdif file[%d]=%s\n", i,infiles[i].c_str());
      exit(1);
    }
    if( in_vdif[i].find_hdr() ) {
      fprintf(stderr,"[%d] %s Hdr-Found:Offset=%d CH:%s\n",i,infiles[i].c_str(),
	      in_vdif[i].Hdr_Offset(), hchstr[ih[i]] );
      edv=in_vdif[i].Hdr().EDV(); 
    }else{
      cerr << "E Failed to find VDIF Header.\n";
      exit(1);
    }
  } // for loop

  /** Sync time and iframe*/
  if(! Sync_frame(in_vdif) ) {
    cerr <<  "E main:Sync_frame(): failed to sync.\n"; exit(1);
  }
  /** Check homogenious data property **/
  if( ! Check_Homo(in_vdif)) { 
    cerr <<  "E main:Check_Homo(): failed checking homogenity of files.\n"; exit(1);
  }

  int  headsize = in_vdif[0].Headsize();;
  int  framesize_in = in_vdif[0].Framesize();
  long datasize_in = framesize_in - headsize;
  long nframe_sec_in = in_vdif[0].Nframe_sec();
  int  qbit = in_vdif[0].Hdr().Qbit();
  int  nch_in = in_vdif[0].Hdr().Nch();
  long in_data_sec = nframe_sec_in * datasize_in;
  long out_data_sec = datasize_in * nframe_sec_in * 8/2;

  if(nch_in < 8) {
    fprintf(stderr,"Error: Nch >8 expected. for this program, but nch=%d, qbit=%d.\n",
	    nch_in, qbit);
    exit(1);
  }

  VDIF_Hdr vhdr_dmy;
  vhdr_dmy.Param_Cpy(in_vdif[0].Hdr());
  if(! stname.empty()) vhdr_dmy.exud_NICT.Set_stname(stname.c_str());
  if(! stid.empty()) vhdr_dmy.Set_Stnid(stid.c_str());

  uint32_t srate = nframe_sec_in * datasize_in * 8/qbit/nch_in;
  vhdr_dmy.Set_Srate(srate/1000); // Set in ksps
  vhdr_dmy.Set_Nch( nch_in * 8/2 ); // Set in ksps
  vhdr_dmy.Set_Epoch( in_vdif[0].Hdr().Epoch() );

  int datasize_out = datasize_in * 4;
  long framesize_out = datasize_out + headsize;
  vhdr_dmy.Set_FrameLen( framesize_out  );
  
  
  VDIF_Frame new_vdif = in_vdif[0];
  if(! stid.empty()) new_vdif.Set_StnID(stid.c_str());
  if(! stname.empty()) new_vdif.Set_StnameID(stname.c_str());

  new_vdif.Set_Sparam(nch_in*4, qbit, srate);
  new_vdif.Set_edv(EDV_NICT);
  RAW_Proto file_prot;
  file_prot.Set_VlbiFrame(&new_vdif);
  file_prot.Set_Host(outfile.c_str());
  if(file_prot.Tx_Ready()){
    printf(" Ready to write outfile(%s) \n", outfile.c_str());
  }else{
    char buf[2048];
    string msg = "E Failed to open outfile: " + outfile + ":";
    msg += strerror_r(errno,buf,2048);
    throw msg;
  }
  
  if(nf_sec ==0) nf_sec= nframe_sec_in;
  new_vdif.Set_Nframe_sec(nf_sec);
  int new_framesize = new_vdif.Framesize();

  fprintf(stderr,"----------------------------------------\n");
  fprintf(stderr,"Framesize: %5d [byte] data/frame: %5ld [byte]", framesize_in, datasize_in);
  fprintf(stderr,"  Nch  : %4d   Qbit : %2d\n", nch_in, qbit);
  fprintf(stderr,"Estimated Frame/sec : %6ld  Data/sec %4ld [MByte/sec]\n", 
	  nframe_sec_in, in_data_sec/1000000 );
  fprintf(stderr,"----------------------------------------\n");
  
  cerr << "========= Conv Elevs2VDIF ver. 1.$Rev: 274 $ start ================\n";
  fprintf(stdout,"=== Converting frame size from %8u Byte (%8u frames/sec) ", 
	  (unsigned int)framesize_in, (unsigned int)in_vdif[0].Nframe_sec());
  fprintf(stdout,"to %8u Byte (%8u frames/sec) \n",
	  (unsigned int)new_framesize, (unsigned int)nf_sec); 
  
  const uint8_t *databuf_out[2];
  for(int i=0; i<2; i++) databuf_out[i] = new uint8_t[out_data_sec];
  uint8_t *dataframe_in[8];
  VDIF_Hdr *phdr[8];
  const uint8_t* dataptr_in[8]; 
  for(int i=0; i<8; i++){ 
    dataframe_in[i]=new uint8_t[framesize_in];
    dataptr_in[i]=dataframe_in[i] + headsize;
    phdr[i]=(VDIF_Hdr *)dataframe_in[i];
    *(phdr[i]) = in_vdif[i].Hdr();
  }

  double total_wsize=0.0;
  long total_wframe=0;
  struct timeval tv0, tv1;  
  gettimeofday(&tv0,0); // Processing start time
  
  long bits_stream_in = qbit*nch_in;
  if( in_vdif[0].Hdr().Data_Type() == VDIF_Hdr::Complex_Sample )  bits_stream_in *= 2;
  int smplB = bits_stream_in/8;
  int hsB8 = smplB*4;
  int hsB = smplB/2;
  int nsmpl_frame= datasize_in/smplB; // Nsample_Frame
  
  size_t wsize=0;
  int tsec=0;
  long cnt=0;
  //  long tcnt=0;
  bool eof_flag=false;
  Time_dat t0 = in_vdif[0].Hdr().Epoch();
  Time_dat t2 ,t1 = t0;
  long nread=1;

  int  ipg = in_vdif[0].Hdr().Sec_Ref() % 2;

  for(int i=0; i<2; i++)  bzero((void* )databuf_out[i], out_data_sec);
  while(eof_flag == false){
    t2=t1;
    for(int ifile=0; ifile < 8; ifile++){
      t1=phdr[ifile]->Epoch();
      cnt=0;
      while(t1 == t2 && nread > 0){
	nread = in_vdif[ifile].Read_Bytes(dataframe_in[ifile], framesize_in);
	cnt++;
	if( edv == phdr[ifile]->EDV() && phdr[ifile]->HD_Match() ){ 
	  if (t1 != phdr[ifile]->Epoch() ){

	    float frame_ratio=(float)cnt/nframe_sec_in * 100.0;
	    fprintf(stdout,"[%d] %s %6ld %6.2f %% \n", ifile,
	    phdr[ifile]->Str_hd(VDIF_Hdr::ONELINE).c_str(), cnt, frame_ratio);
	    t1 = phdr[ifile]->Epoch();
	    cnt=0;
	  }
	  if( nread > 0 ){
	    ipg = phdr[ifile]->Sec_Ref() % 2;
	    int iwpos0= phdr[ifile]->iFrame() * datasize_in*4 + ifile*hsB;
	    uint8_t *rptr = (uint8_t*)dataptr_in[ifile] + ih[ifile] * hsB;
	    uint8_t *wptr = (uint8_t*)databuf_out[ipg] + iwpos0;
	    for(int ismpl=0; ismpl<nsmpl_frame; ismpl++){
	      //	      int iwpos = iwpos0 + ismpl*hsB8;
	      //	      int irpos = ismpl * smplB + ih[ifile] * hsB;
	      //	      const uint8_t* pbuf = dataptr_in[ifile]+ irpos;
	      //	      const uint8_t* wbuf = databuf_out+ iwpos;
	      memcpy((void*)wptr,rptr,hsB); 
	      rptr += smplB;
	      wptr += hsB8;
	    }
	  }else{
	    eof_flag=in_vdif[ifile].EndofFile_Yes();
	  } // if nread
	  //	  if(t1 != t2 )  bzero((void*)dataframe_in[ifile], framesize_in);
	} // If HD_Match()
      } // while t1==t2
      eof_flag=in_vdif[ifile].EndofFile_Yes();
      nread=1;
    } // for i<8
    
    new_vdif.Set_Epoch(t2.Time_t());
    ipg = new_vdif.Hdr().Sec_Ref() % 2;
    wsize = file_prot.TX1sec(t2.Time_t(), databuf_out[ipg], out_data_sec);
    bzero((void*)databuf_out[ipg], out_data_sec);
    total_wsize += (double)wsize;
    total_wframe+= new_vdif.Nframe_sec();
    fprintf(stderr,"wrote  %.0lf MBytes. \n", (total_wsize*1.e-6)); 	  
    
    tsec = t1 - t0; 
    if(lsec > 0 && tsec > lsec ) break;
    
  } // while loop;

  fprintf(stderr," done.\n");
       
  gettimeofday(&tv1,0); // Processing End time
  double dt = (tv1.tv_sec - tv0.tv_sec)+  (tv1.tv_usec - tv0.tv_usec)*1.e-6 ;
  double wrate = total_wsize*1.e-6/dt; // MB/sec
  fprintf(stderr,"wrote  %.1lf MBytes. %ld frames\n", total_wsize*1.e-6, total_wframe); 
  fprintf(stderr," %.3lf sec elapsed(%.1lf MB/s). Finished.\n", dt, wrate);
  cerr << "==========================================\n";

  for(int i=0; i<2; i++)  delete [] databuf_out[i];
  for(int i=0; i<8; i++)  delete [] dataframe_in[i];

}

bool Check_eof(vector <VDIF_Frame>& vdif_arr){
  int nfiles = vdif_arr.size();
  for(int i=0; i<nfiles; i++){
    if( vdif_arr[i].EndofFile_Yes()) return true;
  }
  return false;
}

bool Check_Homo(vector <VDIF_Frame>& vdif_arr){
  int nfiles = vdif_arr.size();
  bool ret = true;
  for(int i=1; i<nfiles; i++){
    if(vdif_arr[0].Framesize() != vdif_arr[i].Framesize()) ret = false;
    if( vdif_arr[0].Hdr().Nch() != vdif_arr[i].Hdr().Nch() ) ret = false;
    if( vdif_arr[0].Hdr().Qbit() != vdif_arr[i].Hdr().Qbit() ) ret = false;
  }
  return ret;
}

bool Sync_frame(vector <VDIF_Frame>& vdif_arr){
  int nfiles = vdif_arr.size();
  int nsync=0;
  uint32_t sec_max(0);
  int iframe_max(0), current_iframe(0);
  unsigned long framesize=vdif_arr[0].Framesize();
  uint8_t *frame_buf = new uint8_t[framesize];
  VDIF_Hdr* phdr = (VDIF_Hdr*)frame_buf;

  while(nsync != nfiles*2 ){
    nsync=0;
    //    if(sec_max != 0) nsync=2;
    for(int ifile=0; ifile < nfiles; ifile++){
      if( sec_max < vdif_arr[ifile].Hdr().Sec_Ref()){
	sec_max = vdif_arr[ifile].Hdr().Sec_Ref();
	iframe_max =ifile;
	current_iframe=vdif_arr[ifile].Hdr().iFrame();
	//	nsync=2;
	continue;
      }
      
      while( sec_max > vdif_arr[ifile].Hdr().Sec_Ref()){
	if ( ! ( vdif_arr[ifile].SeekFrames(1) &&
		 vdif_arr[ifile].Find_FirstFrame()) ){
	  fprintf(stderr, "E main:Sync_frame(): failed at vdif[%d]\n",ifile);
	  return false;
	}
      } // while loop
      if( sec_max == vdif_arr[ifile].Hdr().Sec_Ref() ){
	nsync++;
	int dframe = vdif_arr[iframe_max].Hdr().iFrame() 
	  - vdif_arr[ifile].Hdr().iFrame();
	while(dframe > 0){
	  //	    if( !( vdif_arr[ifile].SeekFrames(dframe) &&
	  //	    if( !( vdif_arr[ifile].SeekFrames(1) &&   vdif_arr[ifile].find_hdr() )){
	  if(  vdif_arr[ifile].Read_Bytes( (void*)frame_buf,framesize) != framesize){
	    fprintf(stderr, "E main:Sync_frame(): failed to read vdif[%d].\n",ifile);
	      return false;
	  }else{
	    vdif_arr[ifile].Hdr() = *phdr;
	    //	    vdif_arr[ifile].Set_Head_par(*phdr);
	  }
	  
	  dframe = vdif_arr[iframe_max].Hdr().iFrame() -   vdif_arr[ifile].Hdr().iFrame();
	}
	if(dframe ==0)	nsync++;
	if(vdif_arr[ifile].Hdr().iFrame()  > current_iframe){
	  iframe_max =  ifile;
	  current_iframe=vdif_arr[iframe_max].Hdr().iFrame();
	}
	
      } // if
    } // for ifile
  } // while loop

  bool ret=true;
  for(int i=0;i<nfiles; i++){
    if( vdif_arr[i].Hdr().Sec_Ref() !=  vdif_arr[iframe_max].Hdr().Sec_Ref() ) ret=false;
    if( vdif_arr[i].Hdr().iFrame()  !=  vdif_arr[iframe_max].Hdr().iFrame() ) ret=false;
  }

  return ret;
}

