/** @file vdif_speana test Spectrum check of VDIF data 
 * @author M.Sekido
 * @Version $Rev: 273 $
 * @Date  $Date: 2017-11-15 10:02:19 +0900 (水, 15 11月 2017) $
 * $Id: vdif_speana.cpp 272 2017-11-15 01:00:12Z vlbi $ 
 */

#include <iostream>
#include <string>
#include <vector>

#include <cstdio>
#include <cstdlib>
#include <algorithm>
#include <functional>

#include "StaticVals.hpp"
#include "VDIF_Header.hpp"
#include "VDIF_Frame.hpp"
#include "DAS_info.hpp"
#include "VsiData.hpp"
#include "pltutl.hpp"
#include "fftw_utl.hpp"
#include "utlty.hpp"
#include "conf_utl.hpp"

const char* version_txt =  "$Rev: 273 $ $Id: vdif_speana.cpp 272 2017-11-15 01:00:12Z vlbi $"; 
extern "C"{
#include <unistd.h>
}

using namespace std;

/** General Configulation data */
Config_set vdif_conf;

const int MAXCH=64;

template <typename T> class dGen{
  /** value to keeping state */
  T r;
  /** delta increment value  */
  T dt;
public:
  /** Default constructor */
  dGen():r(0),dt(1){}
  /** Constructor with initialization */
  dGen(T r0, T d):r(r0),dt(d){}
  /** Setting the initial value */
  void Set_init(const T r0, const T del=0){r = r0; dt=del;}
  /** Setting the delta value */
  void Set_delt(const T d){dt=d;}
  /** Opearator for returning generated value **/ 
  T operator()(){   r += dt; return (r-dt);  }
};

/**! "main_cmn.hpp": Common Declraration of static parameters. 
 * Should be included in main file only. */
#include "main_cmn.hpp"

class mult{
  float f1;
public:
  void Set_fact(float f){f1=f;}
  float operator() (float a){ return a*f1;}
}mult1, mult2;

void usage(){
  cerr << "Usage: " << me << " -f <filename> [-c <nch>] [-t <sec>] [-l]\n";
  cerr << "   [-n <nfft>] [-b <bit encode>] [-g <PGPLOT out>] [-o <outfile>]\n";
  cerr << "   [-i <first ch>] [-e <yyyy.ddd.HH:MM:SS>]\n";
  cerr << "-c <ch>  : Select <ch> data.(default=-1 means all channels.)\n";
  cerr << "-t <sec> : Time duration for integration. (default=0.1 sec.)\n";
  cerr << "-l       : Linear scalse. (default=dB scale 10*log10) \n";
  cerr << "-n <nfft>: Number of FFT points. (default=2048) \n";
  cerr << "-g       : PGPLOT output to file (default=/xw) \n";
  cerr << "-o       : file output of spectrum data.\n";
  cerr << "-i       : first channel to be processed in case total nch > 16.\n";
  cerr << "-e       : Specify epoch of data to be plotted.\n";
  cerr << "-b       : one of [ADS2000/ADS3000/SignMag/RevSign]. (default=ADS2000) \n";
  cerr << "      ADS2000: 00,01,10,11 from bottom to top level.\n";
  cerr << "      ADS3000: 00,10,01,11 from bottom to top level.\n";
  cerr << "      SignMag: 01,00,10,11 from bottom to top level.\n";
  cerr << "      RevSgn : 10,00,01,11 from bottom to top level.\n";
  cerr << "      MS_2bit_1bit   [ml] -> extracting m for 1 bit. \n";
  cerr << "      LS_2bit_1bit   [ml] -> extracting l for 1 bit. \n";
  cerr << "  -----IQ data is detected from VDIF header information-----.\n";
  cerr << "      IQ_1bit: IQ 1bit from LSB to MSB. \n";  
  cerr << "      IQ_2bit: I[ml]Q[ml] 2bit as the same with ADS2000_2bit encoding.\n";
  cerr << " Computing Spectrum for data monitor.\n";
  cerr << " Version: " << version_txt << endl;
  exit(1);
}

/** Averaging data.
 * \return navg Number of averaging count;
 */
const int Avg_spectrum(VDIF_Frame& vdif, const float tdur, const int nfft, 
		       const VsiData::BitEnc_t code_type,
		       vector <float> avg_arr[], DAS_Info &das ){


  /**! Reading Header info  */  
  //  long nsec = 
  //  size_t nsec = 1; // sample 10 sec.

  // First reading data to confirm Srate 
  // cerr << "Reading every sec...";
  // int nframe = vdif.Read_every_sec_hdr(cur_hdrarr, nsec, VDIF_Frame::QUICK);
  // if(nframe < 1){
  //   cerr << "E Failed to find VDIF Header.\n";
  //   exit(1);
  // }
  if(! vdif.find_hdr()){
    cerr << "E Failed to find VDIF Header.\n";
    exit(1);
  }
  long nframe_sec1 = vdif.Nframe_sec();
  
  size_t datasize = vdif.Hdr().Framesize() - vdif.Hdr().Headsize();
  int first_ch = das.FirstCh();
  //  das = vdif.DAS_Inf();// Sampling info
  int nch = vdif.Hdr().Nch();
  int qbit = vdif.Hdr().Qbit();
  long srate=8*datasize/nch/qbit*nframe_sec1;
  das.Set_Param(nch,qbit,srate);
//   if(ch >= nch){
//     fprintf(stderr, "E Avg_spectrum(): Requested ch(%d) >= Nch(%d)\n", ch, nch);
//     return -1;
//   }
  if(nch > MAXCH){
    fprintf(stderr, "E Avg_spectrum(): Nch(%d) in the data > MAXCH(%d)\n", nch, MAXCH);
    return -1;
  }

  long pnt_frame = datasize *8 /qbit/nch; // #points/frame
  //  float sec_frame = (float)datasize/4/das.Speed(); // seconds per frame

  float dt_fft = (float)nfft/(float)srate;  // seconds per fft.
  int navg = (int)(tdur / dt_fft);   // Total #frames requested. 
  if(navg<1){
    fprintf(stderr,"E: Avg_spectrum(): Navg =%d\n", navg);
    return -1;
  }
  //  int tframe = (int)ceilf(tdur/sec_frame);

  VsiData vsidat(&das);
  VsiData::BitEnc_t ecode_type = code_type;
  vector <uint8_t> datbuf;
  vector <FComplex> cmplx_out(nfft/2); // Output FFT Complex 
  vector <float>  amp_arr(nfft/2), phs_arr(nfft/2);
  // , work_arr(nfft/2),work2_arr(nfft/2); 
  if(vdif.Hdr().Data_Type() == VDIF_Hdr::Complex_Sample){
    switch(qbit){
    case 1:  ecode_type = VsiData::IQ_1bit;     break;
    case 2:  ecode_type = VsiData::IQ_2bit;     break;
    }
    cmplx_out.resize(nfft);
    phs_arr.resize(nfft);
    amp_arr.resize(nfft);
    pnt_frame /=2;    
  }
  vsidat.Make_Table(ecode_type); // Make Decode Table 
  
  
  // Number of for reading
  //  nframe = (tdur >1.0 ? (int)(1./sec_frame) : tframe);
  //  nframe = (tframe > 10 ? 10 : tframe);
  //  long npt_read = nframe * pnt_frame;
  //  vector <Fftwf> fftwdat;
  Fftwf fftwdat[MAXCH];

  int npch = nch > 16 ? 16 : nch; // Number of processing channels
  //  int npch = nch;
  const float* pf[npch] ;
  const FComplex* pcf[npch];
  fftwf_complex* pfftw_in[npch];

  //  pf.resize(npch); 
  //  pfftw_in.resize(npch); // all channels 
  
  // Read 1 frame
  long n_read_frm=nfft*nch*qbit/8*1000/datasize;
  vector <VDIF_Hdr> cur_hdrarr;
  ReturnFrameInfo retinfo = vdif.Read_frame(cur_hdrarr, datbuf, n_read_frm);
  if(retinfo.state == ReturnFrameInfo::NG){
    fprintf(stderr, "E Failed to read a frame \n");
    return -1;
  }

 
  long n_available = vsidat.Interpret(datbuf,  nfft*1000);
  if(n_available < 1){ 
    fprintf(stderr,"E Failed VsiDat::Interpret(): ret(%ld) < npt(%ld)\n", 
	    n_available,  pnt_frame);
    return -1;
  }


 
  if(cur_hdrarr[0].Data_Type() == VDIF_Hdr::Real_Sample){
    for(int i=0; i<npch && i+first_ch < nch; i++){ 
      pf[i] = vsidat.pfValArr(i+first_ch); // Pointer to data
      fftwdat[i].Init(nfft);
      pfftw_in[i]  = fftwdat[i].In();
      if(avg_arr[i].size() < (size_t)nfft/2) avg_arr[i].resize(nfft/2);
    }
  }else{ // Complex data
    for(int i=0; i<npch && i+first_ch < nch; i++){
      pcf[i] = vsidat.pcValArr(i+first_ch); // Pointer to data
      fftwdat[i].Init(nfft);
      pfftw_in[i]  = fftwdat[i].In();
      if(avg_arr[i].size() < (size_t)nfft) avg_arr[i].resize(nfft);
    }
  }
  
  long frm_count = 1;
  cerr << "Reading Frame:" << frm_count  << "\r";

  int iavg =0;
  
  while( iavg < navg ){
    for(int i=0; i<npch; i++)  fftwdat[i].Zero_in();

    /** FILL FFT input array **/
    if(nfft < n_available ){
      if(cur_hdrarr[0].Data_Type() == VDIF_Hdr::Real_Sample){ // Real data
	for(int i=0; i<npch; i++){ fftwdat[i].Set_in(pf[i]);  pf[i] += nfft;}
      }else{ // Complex data 
	for(int i=0; i<npch; i++){ fftwdat[i].Set_in(pcf[i]); pcf[i] += nfft;}
      }
      n_available -= nfft;
    }else{ // Number of data is insufficient
      for(int ich=0; ich<npch; ich++){
	pfftw_in[ich]  = fftwdat[ich].In();
	if(cur_hdrarr[0].Data_Type() == VDIF_Hdr::Real_Sample){ // Real data
	  for(int i=0; i<n_available; i++)  pfftw_in[ich][i][0] = pf[ich][i];
	}else{
	  for(int i=0; i<n_available; i++)  memcpy(pfftw_in[ich][i], (void*)&pcf[ich][i],sizeof(FComplex));
	}

	pfftw_in[ich] += n_available;
	pf[ich] += n_available;
      }

      long n_need = nfft - n_available;
      n_available =0;
      
      while(n_need > 0){
	retinfo = vdif.Read_frame(cur_hdrarr, datbuf, n_read_frm);
	if(  retinfo.nframe < 1){
	  cout << endl;
	  break;
	}
	frm_count ++;  // cerr << frm_count << " ";
	n_available = vsidat.Interpret(datbuf,  nfft*1000);
// 	if(n_available < 1){ 
// 	  fprintf(stderr,"E Failed VsiDat::Interpret(): ret(%ld) < npt(%ld)\n",
// 		  n_available, rframe * pnt_frame);
// 	  break;
// 	}
	if(cur_hdrarr[0].Data_Type() == VDIF_Hdr::Real_Sample){ // Real data	
	  for(int ich=0; ich<npch; ich++) pf[ich] = vsidat.pfValArr(ich+first_ch);
	}else{  
	  for(int ich=0; ich<npch; ich++) pcf[ich] = vsidat.pcValArr(ich+first_ch);
	}
	
	if(n_need > n_available){ // In case of sufficient 
	  for(int ich=0; ich<npch; ich++){
	    if(cur_hdrarr[0].Data_Type() == VDIF_Hdr::Real_Sample){ // Real data	
	      for(int i=0; i<n_available; i++)  pfftw_in[ich][i][0] = pf[ich][i];
	      pf[ich] += n_available; 
	    }else{
	      for(int i=0; i<n_available; i++) memcpy(pfftw_in[ich][i],(void*)&pcf[ich][i],sizeof(FComplex));
	      pcf[ich] += n_available; 
	    }
	    pfftw_in[ich] += n_available;
	  }
	  n_need -= n_available; n_available=0;
	}else{  // If enough data is ready
	  for(int ich=0; ich<npch; ich++){
	    if(cur_hdrarr[0].Data_Type() == VDIF_Hdr::Real_Sample){ // Real data	
	      for(int i=0; i<n_need; i++)  pfftw_in[ich][i][0] = pf[ich][i];
	      pf[ich] += n_need; 
	    }else{
	      for(int i=0; i<n_need; i++) memcpy(pfftw_in[ich][i], (void*)&pcf[ich][i],sizeof(FComplex));
	      pcf[ich] += n_need; 
	    }
	    pfftw_in[ich] += n_need;
	  }
	  n_available -= n_need; n_need =0;
	} // else 
      } // End while
    } // else
    /** FILL FFT input array END **/
    
    iavg++;
    float fact1 = 1./(float)iavg;
    float fact2 = (float)(iavg-1)/(float)iavg;
    mult1.Set_fact(fact1);
    mult2.Set_fact(fact2);
    for(int ich=0; ich<npch; ich++){
      fftwdat[ich].Exec(-1);
      //  memcpy(&cmplx_out[0], fftwdat[ich].Out() + nfft/2, sizeof(fftwf_complex)*nfft/2);
      if(vdif.Hdr().Data_Type() == VDIF_Hdr::Real_Sample)
	memcpy(&cmplx_out[0], fftwdat[ich].Out(), sizeof(fftwf_complex)*nfft/2);
      else
	memcpy(&cmplx_out[0], fftwdat[ich].Out(), sizeof(fftwf_complex)*nfft);
      FCarr2AParr(cmplx_out, amp_arr, phs_arr);
      vector <float>::iterator i1 = avg_arr[ich].begin();
      vector <float>::iterator i2 = amp_arr.begin();
      // transform(i1,avg_arr[ich].end(), work_arr.begin(), mult1);
      // transform(i2, amp_arr.end() , work2_arr.begin(), mult2);
      // transform(work_arr.begin(), work_arr.end(), work2_arr.begin(), avg_arr[ich].begin(),
      // 		std::plus<float>());
      for(; i1 < avg_arr[ich].end(); i1++, i2++)  *i1 = *i1*fact2 + *i2*fact1;
    } //  for ich
  } // while tdur 

  fprintf(stderr,"navg=%d %.3lf sec.\n", iavg,(float)iavg*dt_fft);
  return iavg;

}


int main(int argc, char *argv[]){
  me = (me = strrchr(argv[0], '/')) == NULL ? argv[0] : me+1; /* My name */
  enum Plot_t {Linear, dBscale};
  
  string fname;
  int nfft=2048;
  float tdur=0.01;  // (sec)
  VsiData::BitEnc_t  code_type = VsiData::ADS2000_2bit;
  int ch_select = -1; // Channel number 
  Plot_t plot_type = dBscale ; // 10*log10 dB scale
  string pgdev("/xw"), outfile;
  int first_ch=0;
  Time_dat p_epoch; // plot epoch
  bool epoch_flag=false;
  string tmpbuf;
  
  int k=0;
  while (k >= 0) { 
    switch (k = getopt(argc, argv, "hf:c:t:n:b:g:lo:i:e:")) { 
    case 'h':  usage(); break;
    case 'c':    ch_select = strtol(optarg,0,10);  cerr << "  CH=%d is selected\n";  break;
    case 'f':    fname = optarg;      break;
    case 'n':    nfft = strtol(optarg,0,10);      break;
    case 'i':    first_ch = strtol(optarg,0,10);      break;
    case 'l':    plot_type = Linear;  break; // dB scale
    case 't':    tdur = strtof(optarg,0);      break;
    case 'g':    pgdev = optarg;      break;
    case 'o':    outfile = optarg;    break;
    case 'e':    tmpbuf = optarg;  epoch_flag=true; 
      p_epoch.Set_time(tmpbuf,5,":.");  break;
    case 'b': 
      if( strncmp(optarg, "ADS3000",7)==0) code_type = VsiData::ADS3000_2bit;
      if( strncmp(optarg, "SignMag",7)==0) code_type = VsiData::SignMag_2bit;
      if( strncmp(optarg, "RevSign",7)==0) code_type = VsiData::RevSgnMag_2bit;
      if( strncmp(optarg, "MS_2bit_1bit",12)==0) code_type = VsiData::MS_2bit_1bit;
      if( strncmp(optarg, "LS_2bit_1bit",12)==0) code_type = VsiData::LS_2bit_1bit;
      break;
      //    case 'd': vdif_hdr.Set_dbg(atoi(optarg)); break;
    case '?' : /* Error:  character not in list */ 
      fprintf(stderr, /* Yes */ 
	      "%s WARNING:  Unknown option on command line \n", me); 
      usage();
      break; 
    case ':' : /* Error:  missing parameter */ 
      fprintf(stderr, "%s WARNING:  Missing number on command line \n", me); 
      usage();
      break; 
    }
  }

  if(fname.empty()) usage();

  vector <float> avg_arr[MAXCH];
  DAS_Info das;
  
  das.SetFirstCh(first_ch);
  fprintf(stdout,"Nfft=%4d pnts, Integration= %.2f sec. \n", nfft, tdur);
  // Get Averaged Spectrum
  int navg =0;

  VDIF_Frame  vdif;
  if( ! vdif.ROpen(fname)){
    cerr << "Error in opening file: " << fname << endl;
    exit(1);
  }

  if(epoch_flag){
    //    if( vdif.Find_FirstFrame()){
    if( vdif.find_hdr()){
      
      int frmsize= vdif.Framesize();
      long nframe_sec = vdif.Nframe_sec();
      unsigned char *buf = new unsigned char [frmsize];
      VDIF_Hdr* phdr = (VDIF_Hdr*)buf;
      //      while( !vdif.EndofFile_Yes() && vdif.Hdr().Epoch() < p_epoch){
      while( !vdif.EndofFile_Yes() && phdr->Epoch() < p_epoch){
	if(vdif.Hdr().iFrame() < nframe_sec / 2 ) vdif.SeekFrames(nframe_sec/2);
	else vdif.SeekFrames(1);
	//	vdif.find_hdr();
	vdif.Read_Bytes(buf,frmsize);
	vdif.Hdr() = (*phdr);
      }
      if(vdif.EndofFile_Yes()){	cerr << "Eof Found.\n"; exit(1);      }
      fprintf(stderr,"Data Time:%s \n",
	      vdif.Hdr().Epoch().Print_time(14).c_str());
      delete [] buf;
    }else{
      cerr << "E Failed to find First Frame.\n";
      exit (1);
    }

    if( vdif.Hdr().Epoch() != p_epoch)  {
      fprintf(stderr,"E Failed to find epch(%s)\n", 
	    p_epoch.Print_time(14).c_str());
      exit(1);
    }
  }
  
  if( (navg = Avg_spectrum(vdif, tdur, nfft, code_type, avg_arr, das )) < 0){
    cerr << "E Failed in Avg_spectrum() \n";
    exit(1);
  }
  VDIF_Hdr &vdif_hdr = vdif.Hdr();
  fprintf(stdout,"Data Info: %s\n", vdif_hdr.Str_hd(VDIF_Hdr::ALL_ONE).c_str());

  float df = (float)das.Srate()/nfft * 1.e-6; // (MHz)
  vector <float> xarr(nfft/2);
  if(vdif_hdr.Data_Type() == VDIF_Hdr::Complex_Sample)  xarr.resize(nfft);
  dGen<float>g1(0,df);
  generate( xarr.begin(), xarr.end(), g1);

  int nch  = vdif_hdr.Nch();
  int qbit = vdif_hdr.Qbit();

  int npch1 = nch > 16 ? 16 : nch; // Number of processing channels
  // File output
  FILE *ofp=0;
  if( ! outfile.empty()){
    ofp = fopen(outfile.c_str(), "w");
    if(ofp < 0){ cerr << "Failed to open outfile(" << outfile  << ") \n";}
    else{
      fprintf(stderr,"File Output to: %s \n", outfile.c_str());
      fprintf(ofp,"# File: %s (VDIF %s)\n",fname.c_str(),
	      vdif_hdr.Version().c_str());
      fprintf(ofp,"# Mode: %dMsps/%dbit/%dch\n",
	      vdif_hdr.Srate()/1000000,qbit,nch);
      fprintf(ofp,"# Intg:%.2fsec, df=%.2fkHz(n=%d)\n",tdur, df*.1e3, nfft);
    }
  }

  // Graphic Plot of  data 
  Plt_inf pltinf;
  char str1[2][256],str2[256], str3[256];
  sprintf(str1[0],"\\fr%s (VDIF %s)",fname.c_str(),
	  vdif_hdr.Version().c_str());
  //  if(hdrarr[0].ext_dat_ver == 1) // Extension NICT version
  sprintf(str1[1],"\\fr%dMsps/%dbit/%dch",
	  vdif_hdr.Srate()/1000000,qbit,nch);
  sprintf(str2,"\\frIntg:%.2fsec, df=%.2fkHz(n=%d)",tdur, df*.1e3, nfft);
  cpgopen( pgdev.c_str() ); 
  switch(plot_type){
  case Linear:   pltinf.Set_label("\\frFreq. (MHz)", "\\frPower"); break;
  case dBscale:
  default:
    pltinf.Set_label("\\frFreq. (MHz)", "\\frPower [dB]"); 
//     pltinf.Set_ymax(30);
//     pltinf.Set_ymin(-10);
//     pltinf.Set_ytick(10);
    break;
  }

  
  int npch =1; // Number ploting channel 
  if(ch_select == -1)  npch = npch1; 
  switch(npch){
  case 2:
    cpgsubp(2,1);
    break;
  case 4:
    cpgsubp(2,2);
    break;
  case 8:
    cpgsubp(4,2);
    break;
  case 16:
  case 32:
  case 64:
    cpgsubp(4,4);
    break;
  case 1:
  default:
    break;
  }
  cpgask(1);
  vector <float> dbarr, yarr(xarr.size());

  for(int ich=0; ich<npch; ich++) {
    if(npch >= 16 && ich > 0 && ich % 16 ==0) cpgask(1);
    else cpgask(0);
    cpgpage();
    cpgsvp(0.15, 0.9, 0.15, 0.9);  

    if(ch_select != -1) // Specific Channel is selected 
      ich = ch_select;
    if(ofp > 0) fprintf(ofp,"# CH=%d \n", ich+first_ch);
    sprintf(str3,"\\frCH=%d", ich+first_ch);

    if(plot_type == Linear){  
      copy(avg_arr[ich].begin(), avg_arr[ich].end(),yarr.begin());
    }else {
      yarr = val2dB(avg_arr[ich]);
	//	pltinf.Plot1(xarr,dbarr);
    }

    pltinf.Plot1(xarr,yarr);
    if(ch_select != -1)  ich = ch_select; // Exit loop in case of specific channel.
    sprintf(str3,"\\frCH=%d", ich+first_ch);
    //    cpgsch(1.5);
    cpgsch(2.5);
    cpgmtxt("T", -1, 0.9,1,str3);

    if(ofp > 0){ /// File output
      for(uint i=0; i<xarr.size(); i++) 
	fprintf(ofp,"%7.3f  %+12.4e\n", xarr[i], yarr[i]);
    }
    
  }
  cpgask(1);
  cpgsubp(1,1);
  cpgsch(0.8);
  cpgsvp(0.1,0.9,0.1,0.9);
  //  if(ich==0){
  cpgmtxt("T", 4, 0,0, str1[0]);
  cpgmtxt("T", 4, 0.9,1, str1[1]);
  cpgmtxt("T", 2, 0., 0, str2);
  //  }

  //  cpgpage();
  cpgclos();
  if(ofp > 0) fclose(ofp); /// File output

}
