
#ifndef __DATAPAGE_HPP__
#define __DATAPAGE_HPP__
/** Data Page (Memory) device class library
 * @Version $Rev: 142 $
 * @Author M.Sekido
 * @Date  $Date: 2014-10-17 09:11:23 +0900 (Fri, 17 Oct 2014) $
 * $Id: DataPage.hpp 142 2014-10-17 00:11:23Z vlbi $ 
 */

#include <vector>
#include <iostream>
#include <algorithm>
using namespace std;

#include "Mutex.hpp"
#include "Logsystem.hpp"

extern "C"{
#include <time.h>
#include <stdint.h>
#include <unistd.h>
}

const uint32_t DEFAULT_PAGE_SIZE = 8000000;
const int      DEFAULT_NPAGE     = 6;
const int      N_LOOP_LIMIT      = 20;  // (sec)

/** Data Buffer class with time tag
 */
class DataPage{
public:
  /** Data page status types
   *    - RW_OK means the data is available and Writable
   *    - RD_OK means the data is available and Read Only
   *    - RPROTECTED means the data is being written and  not ready.
   *    - EMPTY means the data is not ready and writable.
   *    - DATA_END  means the last page.
   */
  enum PageState_t {WR_OK, RD_OK, R_PROTECTED, EMPTY, DATA_END };
  //  enum WaitMode_t {WAIT, NONWAIT};
  DataPage(); ///< Constructor
  DataPage(const int npg); ///< Constructor
  ~DataPage(); ///< Destructor
  /// return reference of vector array of i page
  vector <uint8_t>& PagePoint(const int i);
  //  void Set_Npage(const int n);
  const uint32_t Pagesize() const {return pagesize;} ///< Returns page size in bytes
  const int PageResize(const uint32_t pgsz); ///< Update Pagesize
  const time_t Epoch(const int i)const ; ///< Return epoch of i page;
  const int Npage()const {return npage;} ///< Returns Number of pages
  time_t EarliestEpoch(); ///< Returns earliest epoch in the all pages
  /// Returns Page number of currently writing
  int IpageWriting()const {return ipage_writing;}
  /** Step writing one page forward with writing page state 'state_dat', 
   *  The state is left with 'R-Protected', Mainly used to write data to the page,
   *  But this is only one method to update ipage_wriring parameter.
   */
  int WStepPage(const PageState_t state_dat, const bool wmode=true); 
  /// Chaging Page state of ipage to 'state_dat' immediately, but not update ipage_writing.
  int ChgPageState(const int ipage, const PageState_t state_dat);
  /// Returns reference of vector array of writing page.
  vector <uint8_t>& WritingPagePtr() {return dat[ipage_writing];}
  /// Set time to the page currently writing data 
  void SetTimeWritingPage(const time_t t1);
  void PrintPageState(); ///< Print Page state
  string StrPageState(const int lvl=0); ///< Return Page state in string
  void InitPages(); ///< Initialize all pages
  const bool Normal()const; ///< Return true if status is normal
  const bool TestPage(const int ipg, const PageState_t t)const; 
  void PageEnd();
  void MutexRelease(){mutex.unlock();}
  const PageState_t PageState(const int ipage) const {return status[ipage];}
  const uint8_t* Data(const int ipg, const uint32_t pos=0) const{ 
    return &dat[ipg][pos];}
  const time_t Time_t(const int ipg) const {return tarr[ipg];}
private:
  uint32_t pagesize; ///< Pagesize in bytes
  const int npage; ///< Number of pages
  int ipage_writing; ///< Protected page;
  vector <uint8_t> dat[DEFAULT_NPAGE]; ///< array of data page
  vector <time_t>  tarr;  ///< time array 
  vector <PageState_t>  status; ///< status array
  Mutex mutex; ///< mutex instance
  //  uint32_t wait_limit; ///< page Waiting limit (-1: endless)
};

inline string PgState2Str(const DataPage::PageState_t s){
  string ret;
  switch(s){
  case DataPage::RD_OK:       ret = "RD_OK "; break;
  case DataPage::WR_OK:       ret = "WR_OK "; break;
  case DataPage::R_PROTECTED: ret = "R_PRT "; break;
  case DataPage::EMPTY:       ret = "EMPTY "; break;
  case DataPage::DATA_END:    ret = "DATEND"; break;
  }
  return ret;
}

inline string Epoch2Str(const time_t t){
  char dumbuf[256];
  struct tm tm1;
  strftime(dumbuf,256,"%Y/%j-%H:%M:%S",  gmtime_r(&t, &tm1));
  string ret = dumbuf;
  return ret;
}

inline void DataPage::SetTimeWritingPage(const time_t t1){
  mutex.lock();
  tarr[ipage_writing]=t1;
  mutex.unlock();
}

/**
 * Changing page state regardless to the current page state.
 *  input:
 *    - ipage: Page Number 
 *    - state_dat: State data to be witten to the page.
 *  This method is only used for intialization or releasing the page state.
 */
inline int 
DataPage::ChgPageState(const int ipage, const PageState_t state_dat){
  mutex.lock();
  //  cerr << "ChngPageState " << ipage << ":";  PrintPageState();
  Logsystem::debug_logger()->Tstamp() 
    << "ChgPageState(): (" << ipage << "pg)" << PgState2Str(status[ipage]) << "("
    << Epoch2Str(tarr[ipage]) << ") -> " << PgState2Str(state_dat) << "\n";
  Logsystem::debug_logger()->Flush();
  status[ipage] = state_dat;
  mutex.unlock();
  return 0;
}


inline const time_t DataPage::Epoch(const int i) const{
  if( i >= (int)tarr.size() ){
    cerr << "E DataPage::Epoch(): ipage(" << i << ") > pagesize(" 
	 << tarr.size() << ")\n";
    return 0;
  }
  return tarr[i];
}

inline const bool DataPage::Normal()const{

  for(int i=0; i< (int)status.size(); i++){
    if(status[i] == DATA_END) return false;
  }
  return true;
  
}

inline void DataPage::PageEnd(){
  mutex.lock();
  fill(status.begin(), status.end(),DataPage::DATA_END);
  mutex.unlock();
}
inline const bool 
DataPage::TestPage(const int ipg, const PageState_t t) const  { 
  if(ipg > (int)status.size() || ipg < 0) return false; 
  return status[ipg] == t ;
}

#endif
