
#ifndef __VSIDATA_HPP__
#define __VSIDATA_HPP__
/** 
 * A Class for VSI data Handling
 * @Version $Rev: 235 $
 * @Author M.Sekido
 * @Date  $Date: 2017-09-18 23:14:38 +0900 (月, 18  9月 2017) $
 * $Id: VsiData.hpp 235 2017-09-18 14:14:38Z vlbi $ 
 */

#include <iostream>
#include <vector>
using namespace std;
#include "DAS_info.hpp"
#include "cmplx.hpp"

extern "C"{
#include <stdint.h>
}

const int MAX_NBIT = 8;
// const int MAX_NCH = 8;

#ifndef MAX_NCH
#define MAX_NCH (64)
#endif

 
class VsiData{
public:
  /**! Bit encoding order, Same with ADS2000 or its reverse. 
   * ADS2000_2bit-> 00, 01, 10, 11, from bottom to the top level
   * ADS3000_2bit-> 00, 10, 01, 11, from bottom to the top level
   * SignMag_2bit     -> 01, 00, 10, 11, from bottom to the top level
   * RevSignMag_2bit  -> 10, 00, 01, 11, from bottom to the top level
   * IQ_1bit          -> IQ 1bit from LSB to MSB  
   * IQ_2bit          -> I[ml]Q[ml] 2bit as the same with ADS2000_2bit encoding.
   * MS_2bit_1bit   [ml] -> extracting m for 1 bit
   * LS_2bit_1bit   [ml] -> extracting l for 1 bit
   */
  enum SampleType_t  {Real_Sample=0, IQ_Sample=1};
  enum BitEnc_t {ADS2000_2bit, ADS3000_2bit, SignMag_2bit, RevSgnMag_2bit,
		 MS_2bit_1bit, LS_2bit_1bit, IQ_1bit, IQ_2bit};
  /** Constructor */
  VsiData(DAS_Info* das=0):_s_type(Real_Sample),_dasInfo(das){
    for(int i=0; i<MAX_NCH; i++) _fcarr[i]=0;}
  ~VsiData(){ for(int i=0; i<MAX_NCH; i++) 
      if(_fcarr[i] !=0){ delete [] _fcarr[i]; _fcarr[i]=0;}}
  /** Making look-up table. */
  int Make_Table(const BitEnc_t t=ADS2000_2bit);
  /** Set DAS information class */
  void Set_DasInfo( DAS_Info* das){ _dasInfo = das;  }
  /** Reading 1 sec of data from the file */
  const int32_t Read1Sec(FILE *fp);
  /** Value of the data at (ch,ith) */
  const int8_t Val(const int ch, const int32_t i);
  /** Interpretation (decoding)
   * \return Number of data points.   */
  const long Interpret( const long npt=0);
  /** Interpretation (decoding) 
   * \param datbuf Binary Byte date
   * \param npt Number of data to be converted
   * \param tgt_type Conversion data type 0:Flat, 1:Integer
   * \return Number of data points.   */
  const long Interpret(vector <uint8_t>& datbuf, const long npt, const int tgt_type=0);
  /** Returns reference of value array at ch*/
  vector <int8_t>& ValArr(const int ch){ return _valarr[ch];}
  /** Returns pointer of value array at ch */
  const int8_t *pValArr(const int ch) {return &_valarr[ch][0];} ///< Returns pointer of one
  /** Returns reference of float value array at ch*/
  vector <float>& fValArr(const int ch){ return _fvalarr[ch];}
  /** Returns pointer of float value array at ch */
  const float *pfValArr(const int ch) {return &_fvalarr[ch][0];} ///< Returns pointer of one
  /** Returns Reference of FComplex array at ch */
  //  vector <FComplex>& cValArr(const int ch) {return _fcarr[ch];} ///< Returns pointer of one
  FComplex * cValArr(const int ch) {return _fcarr[ch];} ///< Returns pointer of one
  /** Returns pointer of FComplex array at ch */
  //  const FComplex *pcValArr(const int ch) {return &_fcarr[ch][0];} ///< Returns pointer of one
  const FComplex *pcValArr(const int ch) {return _fcarr[ch];} ///< Returns pointer of one
private:
  const int MkTbl1bit(void);
  const int MkTbl2bit(const BitEnc_t t=ADS2000_2bit);
  const int MkTbl4bit(void);
  const int MkTbl8bit(void);
  const int Conv_data_d_unitbyte(vector <uint8_t>& datbuf, const int nch, 
				 const int unit_byte, const  size_t np);
  const int Conv_data_f_unitbyte(vector <uint8_t>& datbuf, const int nch, 
				 const int unit_byte, const  size_t np);
  const int Conv_data_c_unitbyte(vector <uint8_t>& datbuf, const int nch, 
				 const int unit_byte, const  size_t np);
  const int Conv_data_d_byteunit(vector <uint8_t>& datbuf, const int nch, 
				 const int byte_unit, const  size_t np);
  const int Conv_data_d_byteunit(vector <uint8_t>& datbuf, const int nch, 
					  const int byte_unit, const int ch_byte, const  size_t npt );
  const int Conv_data_f_byteunit(vector <uint8_t>& datbuf, const int nch, 
					  const int byte_unit, const int ch_byte, const  size_t npt );
  const int Conv_data_c_byteunit(vector <uint8_t>& datbuf, const int nch, 
					  const int byte_unit, const int ch_byte, const  size_t npt );
  SampleType_t  _s_type; 
  DAS_Info *_dasInfo;
  int8_t _table[256][MAX_NBIT];
  float  _ftable[256][MAX_NBIT];
  FComplex _fctable[256][MAX_NBIT];
  FComplex *_fcarr[MAX_NCH];
  vector <uint8_t> _buf;
  vector <int8_t> _valarr[MAX_NCH]; 
  vector <float> _fvalarr[MAX_NCH]; 
  //  vector <FComplex> _fcarr[MAX_NCH];
};

const int8_t level_1bit[2] = {-1,1};
const int8_t level_2bit[4] = {-3, -1, 1, 3 };
const int8_t level_ADS3000_2bit[4] = {-3, 1, -1, 3 };
const int8_t level_SignMag_2bit[4] = {1, -3, 1, 3 };
const int8_t level_RevSgnMag_2bit[4] = {-1, 1, -3, 3};


inline const int8_t VsiData::Val(const int ch, const int32_t ith){
  if(_dasInfo ==0){  
    cerr << "Error @VsiData::Val():dasInfo is empty.\n";
    return -100;  
  }
  int nch= _dasInfo->Nch();
  int bit = _dasInfo->Qbit();
  int dunit = nch * bit;
  int ch_byte = 8/bit;
  int unit_byte(0);
  int byte_unit (0);
  char msg[256];

  try{
    if( nch > ch_byte){  // 1 unit takes more than 1 ch
      if( dunit % 8 != 0){
	sprintf(msg, "Nch(%d)*Qbit(%d) must be multiple of 8 bit or less than 8bit", nch, bit);
	throw msg;
      }
      byte_unit = dunit / 8;
    }else{  // 1 byte include no less than 1 channel
      if(8 % dunit != 0){
	sprintf(msg, "Nch(%d)*Qbit(%d) must be multiple of 8 bit or less than 8bit", nch, bit);
	throw msg;
      }else{
	unit_byte = 8/dunit;
      }
    }
  } 
  catch(const char* msg){
    cerr << "Error in VsiData::Val():" << msg << endl;
    return -100;
  }

  if( byte_unit == 0 ){
    int32_t ibyte = ith / unit_byte;
    int frac_unit = ith %  unit_byte;
    return _table[_buf[ibyte]][ch+frac_unit*nch];
  }else{
    int32_t ibyte = ith*byte_unit;
    int abyte = ch /ch_byte;
    ibyte += abyte;
    int ich = ch % ch_byte;
    return _table[_buf[ibyte]][ich];
  }
  
}

// template <typename T>
// const int Conv_data_s(vector <uint8_t>& datbuf, const int nch, 
// 		      vector <T> varr[MAX_NCH], const T table[][MAX_NBIT], 
// 		      const int unit_byte, const  size_t npt){
//   vector <uint8_t>::iterator sptr = datbuf.begin();  
//   vector <T>::iterator ptr[MAX_NCH];
//   for(int i=0; i<nch; i++){ ptr[i] = varr[i].begin();   }

//   uint32_t cnt=0;
//   while(sptr < datbuf.end() && ptr[0] < varr[0].end() && cnt < npt){
//     for( int i=0; i< unit_byte; i++){
//       for(int j=0; j<nch; j++){
// 	*ptr[j] = table[*sptr][j+i*nch]; 
// 	ptr[j]++; 
// 	//	    dptr2[j]++;
//       }
//       cnt++;
//     }
//     sptr++;
//   } 
// }


// template <typename T>
// const int Conv_data(vector <uint8_t>& datbuf, const int nch, 
// 		    vector <T> varr[], const T table[][MAX_NBIT], 
// 		    const int byte_unit, const int ch_byte, const  size_t npt ){
  
//       vector <uint8_t>::iterator sptr = datbuf.begin();  
//       while(sptr < datbuf.end() && ptr[0] < _fvalarr[0].end() && cnt < npt){
// 	int  ich=0;
// 	for(int i=0; i<byte_unit; i++){ // bytes/unit
// 	  for(int jch=0; jch< ch_byte; jch++,ich++){ // ch/byte
// 	    *dptr[ich] = _ftable[*sptr][jch];  
// 	    *dptr2[ich] = _table[*sptr][jch];  
// 	    dptr[ich]++;   dptr2[ich]++; 
// 	  }
// 	  sptr++;
// 	}
// 	cnt++;
//       }
//     } 

// }

#endif
